@@ -3,10 +3,9 @@ package Dancer2::Plugin::Auth::Extensible::Provider::DBIC;
 use strict;
 use base 'Dancer2::Plugin::Auth::Extensible::Provider::Base';
 use DBIx::Class::ResultClass::HashRefInflator;
-use Lingua::EN::Inflect::Phrase;
 use String::CamelCase qw(camelize);
 
-our $VERSION = '0.401';
+our $VERSION = '0.500';
 
 =head1 NAME 
 
@@ -66,11 +65,16 @@ A full example showing all options:
                 users:
                     provider: 'DBIC'
 
-                    # optionally specify names of tables if they're not the defaults
-                    # (defaults are 'users', 'roles' and 'user_roles')
-                    users_table: 'users'
-                    roles_table: 'roles'
-                    user_roles_table: 'user_roles'
+                    # Optionally specify the sources of the data if not the
+                    # defaults (as shown).  See notes below for how these
+                    # generate the resultset names.  If you use standard DBIC
+                    # resultset names, then these and the column names are the
+                    # only settings you might need.  The relationships between
+                    # these resultsets is automatically introspected by
+                    # inspection of the schema.
+                    users_source: 'user'
+                    roles_source: 'role'
+                    user_roles_source: 'user_role'
 
                     # optionally set the column names
                     users_username_column: username
@@ -83,8 +87,8 @@ A full example showing all options:
 
                     # Optionally set columns for user_password functionality in
                     # Dancer2::Plugin::Auth::Extensible
-                    users_pwresetcode_column:   # Reset code column. No default.
-                    users_pwchanged_column:     # Time of reset column. No default.
+                    users_pwresetcode_column: pw_reset_code
+                    users_pwchanged_column:   # Time of reset column. No default.
 
                     # Days after which passwords expire. See logged_in_user_password_expired
                     # functionality in Dancer2::Plugin::Auth::Extensible
@@ -108,22 +112,39 @@ A full example showing all options:
                     # Optionally specify the algorithm when encrypting new passwords
                     encryption_algorithm: SHA-512
 
+                    # If you don't use standard DBIC resultset names, you might
+                    # need to configure these instead:
+                    users_resultset: User
+                    roles_resultset: Role
+                    user_roles_resultset: UserRole
+
+                    # Deprecated settings. The following settings were renamed for clarity
+                    # to the *_source settings, although they can still be used. There is
+                    # no plan to remove them.
+                    users_table:
+                    roles_table:
+                    user_roles_table:
+
+
 =over
 
-=item users_table
+=item user_source
 
-Specifies the database table that the users are stored in. This will be camelized.
+Specifies the source name that contains the users. This will be camelized to generate
+the resultset name. The relationship to user_roles_source will be introspected from
+the schema.
 
-=item roles_table
+=item role_source
 
-Specifies the database table that the roles are stored in. This will be camelized.
+Specifies the source name that contains the roles. This will be camelized to generate
+the resultset name. The relationship to user_roles_source will be introspected from
+the schema.
 
-=item user_roles_table
+=item user_roles_source
 
-Specifies the database table that holds the many-to-many relationship information
-between users and roles. It is assumed that the relationship is configured in
-the DBIC schema, such that a user has many entries in the user_roles table, and
-that each of those has one role. This table name will be pluralized.
+Specifies the source name that contains the user_roles joining table. This will be
+camelized to generate the resultset name. The relationship to the user and role
+source will be introspected from the schema.
 
 =item users_username_column
 
@@ -173,10 +194,24 @@ keyword. Instead it is intended to make it easier to access a user's roles if th
 user hash is being passed around (without requiring access to the user_has_role
 keyword in other modules).
 
+=item users_resultset
+
+=item roles_resultset
+
+=item user_roles_resultset
+
+These configuration values are provided for fine-grain tuning of your DBIC
+resultset names. If you use standard DBIC naming practices, you will not need
+to configure these, and they will be generated internally automatically.
+
+=back
+
 =head1 SUGGESTED SCHEMA
 
-See the L<Dancer2::Plugin::Auth::Extensible::Provider::Database> documentation
-for an example schema.
+Please see Schema1 in the tests directory for a suggested schema.
+
+If producing a schema from scratch, it is recommended that the default
+resultset and column names are used, as per the default configuration.
 
 =cut
 
@@ -184,22 +219,63 @@ for an example schema.
 sub new {
     my ($class, $realm_settings, $dsl) = @_;
 
+    if ( $dsl->app->can('with_plugin') ) {
+        # plugin2
+        $dsl->app->with_plugin('Dancer2::Plugin::DBIC')
+          or die "Failed to load Dancer2::Plugin::DBIC";
+    }
+    else {
+        # old bad world
+        die "No schema method in app. Did you load Dancer2::Plugin::DBIC before Dancer2::Plugin::Auth::Extensible?"
+          unless $dsl->can('schema');
+    }
+
     # Grab a handle to the Plugin::DBIC schema
-    die "No schema method in app. Did you load DBIC::Plugin::DBIC before DBIC::Plugin::Auth::Extensible?"
-        unless $dsl->can('schema');
     my $schema = $realm_settings->{schema_name}
                ? $dsl->schema($realm_settings->{schema_name})
                : $dsl->schema;
 
     # Set default values
-    $realm_settings->{users_table}            ||= 'users';
-    $realm_settings->{users_username_column}  ||= 'username';
-    $realm_settings->{users_lastlogin_column} ||= 'lastlogin';
-    $realm_settings->{user_valid_conditions}  ||= {};
-    $realm_settings->{users_password_column}  ||= 'password';
-    $realm_settings->{roles_table}            ||= 'roles';
-    $realm_settings->{user_roles_table}       ||= 'user_roles';
-    $realm_settings->{roles_role_column}      ||= 'role';
+    $realm_settings->{users_source}                 ||= ($realm_settings->{users_table} || 'user');
+    $realm_settings->{users_resultset}              ||= camelize($realm_settings->{users_source});
+    $realm_settings->{users_username_column}        ||= 'username';
+    $realm_settings->{users_lastlogin_column}       ||= 'lastlogin';
+    $realm_settings->{user_valid_conditions}        ||= {};
+    $realm_settings->{users_password_column}        ||= 'password';
+    $realm_settings->{roles_source}                 ||= ($realm_settings->{roles_table} || 'role');
+    $realm_settings->{roles_resultset}              ||= camelize($realm_settings->{roles_source});
+    $realm_settings->{user_roles_source}            ||= ($realm_settings->{user_roles_table} || 'user_roles');
+    $realm_settings->{user_roles_resultset}         ||= camelize($realm_settings->{user_roles_source});
+    $realm_settings->{roles_role_column}            ||= 'role';
+    $realm_settings->{users_pwresetcode_column}     ||= 'pw_reset_code';
+
+    # introspect result sources to find relationships
+
+    my $user_roles_class =
+      $schema->resultset( $realm_settings->{user_roles_resultset} )
+      ->result_source->result_class;
+
+    foreach my $name (qw/user role/) {
+
+        my $result_source =
+          $schema->resultset( $realm_settings->{"${name}s_resultset"} )
+          ->result_source;
+
+        foreach my $relname ( $result_source->relationships ) {
+            my $info = $result_source->relationship_info($relname);
+            my %cond = %{ $info->{cond} };
+            if (   $info->{class} eq $user_roles_class
+                && $info->{attrs}->{accessor} eq 'multi'
+                && $info->{attrs}->{join_type} eq 'LEFT'
+                && scalar keys %cond == 1 )
+            {
+                $realm_settings->{"${name}_user_roles_relationship"} = $relname;
+                ( $realm_settings->{"${name}_relationship"} ) =
+                  keys %{ $result_source->reverse_relationship_info($relname) };
+                last;
+            }
+        }
+    }
 
     my $self = {
         realm_settings => $realm_settings,
@@ -213,7 +289,6 @@ sub new {
 sub _user_rset {
     my ($self, $column, $value, $options) = @_;
     my $settings              = $self->realm_settings;
-    my $users_table           = $settings->{users_table};
     my $username_column       = $settings->{users_username_column};
     my $user_valid_conditions = $settings->{user_valid_conditions};
 
@@ -228,7 +303,7 @@ sub _user_rset {
     my $search = { %$user_valid_conditions, $search_column => $value };
 
     # Look up the user
-    $self->_schema->resultset(camelize $users_table)->search($search, $options);
+    $self->_schema->resultset($settings->{users_resultset})->search($search, $options);
 }
 
 sub _dsl_local { shift->{dsl_local} };
@@ -327,11 +402,10 @@ sub get_user_by_code {
 sub create_user {
     my ($self, %user) = @_;
     my $settings        = $self->realm_settings;
-    my $users_table     = $settings->{users_table};
     my $username_column = $settings->{users_username_column};
-    my $username        = $user{username}
+    my $username        = delete $user{username} # Prevent attempt to update wrong key
         or die "Username needs to be specified for create_user";
-    $self->_schema->resultset(camelize $users_table)->create({
+    $self->_schema->resultset($settings->{users_resultset})->create({
         $username_column => $username
     });
     $self->set_user_details($username, %user);
@@ -354,34 +428,31 @@ sub set_user_details {
     if (my $roles_key = $self->realm_settings->{roles_key}) {
         if (my $new_roles = delete $update{$roles_key}) {
 
-            my $users_table           = $settings->{users_table};
-            my $roles_table           = $settings->{roles_table};
-            my $user_roles_table      = $settings->{user_roles_table};
             my $roles_role_column     = $settings->{roles_role_column};
             my $users_username_column = $settings->{users_username_column};
 
-            my @all_roles      = $self->_schema->resultset(camelize $roles_table)->all;
+            my @all_roles      = $self->_schema->resultset($settings->{roles_resultset})->all;
             my %existing_roles = map { $_ => 1 } @{$self->get_user_roles($username)};
 
             foreach my $role (@all_roles) {
-                my $role_name = $role->name;
+                my $role_name = $role->$roles_role_column;
                 if ($new_roles->{$role_name} && !$existing_roles{$role_name}) {
                     # Needs to be added
-                    $self->_schema->resultset(camelize $user_roles_table)->create({
-                        $users_table => {
+                    $self->_schema->resultset($settings->{user_roles_resultset})->create({
+                        $settings->{user_relationship} => {
                             $users_username_column => $username,
                             %{$settings->{user_valid_conditions}}
                         },
-                        $roles_table => { $roles_role_column => $role_name },
+                        $settings->{role_relationship} => { $roles_role_column => $role_name },
                     });
                 }
                 elsif (!$new_roles->{$role_name} && $existing_roles{$role_name}) {
                     # Needs to be removed
-                    $self->_schema->resultset(camelize $user_roles_table)->search({
-                        "$users_table.$users_username_column" => $username,
-                        "$roles_table.$roles_role_column"     => $role_name,
+                    $self->_schema->resultset($settings->{user_roles_resultset})->search({
+                        "$settings->{user_relationship}.$users_username_column" => $username,
+                        "$settings->{role_relationship}.$roles_role_column"     => $role_name,
                     },{
-                        join => [ $users_table, $roles_table ],
+                        join => [ $settings->{user_relationship}, $settings->{role_relationship} ],
                     })->delete;
                 }
             }
@@ -404,13 +475,12 @@ sub set_user_details {
 sub get_user_roles {
     my ($self, $username) = @_;
 
-    my $settings          = $self->realm_settings;
-    my $roles_table       = $settings->{roles_table};
-    my $user_roles_table  = $settings->{user_roles_table};
-    my $roles_role_column = $settings->{roles_role_column};
-    $user_roles_table     = Lingua::EN::Inflect::Phrase::to_PL($user_roles_table);
+    my $settings                     = $self->realm_settings;
+    my $role_relationship            = $settings->{role_relationship};
+    my $user_user_roles_relationship = $settings->{user_user_roles_relationship};
+    my $roles_role_column            = $settings->{roles_role_column};
 
-    my $options = { prefetch => { $user_roles_table => $roles_table } };
+    my $options = { prefetch => { $user_user_roles_relationship => $role_relationship } };
     my ($user) = $self->_user_rset(username => $username, $options)->all;
     if (!$user) {
         $self->_dsl_local->debug("No such user $username when looking for roles");
@@ -418,9 +488,9 @@ sub get_user_roles {
     }
 
     my @roles;
-    foreach my $ur ($user->$user_roles_table)
+    foreach my $ur ($user->$user_user_roles_relationship)
     {
-        my $role = $ur->$roles_table->$roles_role_column;
+        my $role = $ur->$role_relationship->$roles_role_column;
         push @roles, $role;
     }
 
